function [x, ind] = foba(A, y, Kmax, varargin)
% Forward-backward pursuit, from the paper: N. B. Karahanoglu, H. Erdogan,
% Forward-backward search for compressed sensing signal recovery, EUSIPCO
% 2012.
%
% Usage: x = foba(A, y, varargin)
%
% Inputs: A - system matrix
%
%         y - measurement vector
%
%         Kmax - maximal number of nonzero elements of the solution
%
%         optional parameters: alpha - number of indexes to select in every
%         iteration
%
%                              beta - number of indexes to remove (beta <
%                              alpha)
%
%                              gram - precomputer Gram matrix (default:
%                              false)
%
%                              impl - implementation (default: cholesky)
%
%                              epsilon - tolerance on the residual
%                              (default: eps)
%
%                              verbose - show progress (default: false)
%
% (input opt. parameters as ('name', 'value') pairs)
%
% Outputs: x - approximation of the solution
%
%          ind - support of x
%
% Marko Filipovic, 03.-12.12.2012.

alpha = 2;

epsilon = eps;
% Implementation: naive, cholesky, qr
implementation = 'cholesky';
precompute_gram = false;
maxiter = size(A, 2);
verbose = false;

% Optional parameters
% Check for optional parameters
if(mod(length(varargin),2) == 1)
    error('Optional arguments come in pairs.');
end
for i = 1:2:length(varargin)
    switch lower(varargin{i})                
        case 'alpha'
            alpha = varargin{i+1};  
        case 'beta'
            beta = varargin{i+1};  
        case 'gram'
            precompute_gram = varargin{i+1};
        case 'impl'
            implementation = varargin{i+1};
        case 'epsilon'
            epsilon = varargin{i+1};
        case 'verbose'
            verbose = varargin{i+1};
    end
end

if alpha >= size(A, 2)
    error('Alpha is too big')
end

if ~exist('beta', 'var')
    beta = alpha - 1;
end
if beta >= alpha
    beta = alpha - 1;
end

% Initial residual
res = y;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if verbose
    hh = waitbar('FoBA in progress...');
end

switch implementation
    
    case 'cholesky'
        
        % Precompute A' * y
        Aty = A' * y;

        if precompute_gram
            G = A' * A;
        end

        % Options for linsolve
        linsolve_opts.LT = true;
        linsolve_opts_transpose.LT = true;
        linsolve_opts_transpose.TRANSA = true;
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        [~, ind] = sort(abs(A' * res), 'descend');
        ind = ind(1:alpha);

        % Initial Cholesky factor
        if precompute_gram
            CholFact = chol(G(ind, ind), 'lower');
        else
            CholFact = chol(A(:, ind)' * A(:, ind), 'lower');
        end

        % Update the approximation
        xt = linsolve(CholFact, Aty(ind), linsolve_opts);
        xt = linsolve(CholFact, xt, linsolve_opts_transpose);

        [~, ind_rem] = sort(abs(xt));        
        ind_preserve = ind_rem(beta+1:end);
        
        ind = ind(ind_preserve);
        
        if precompute_gram
            CholFact = chol(G(ind, ind), 'lower');
        else
            CholFact = chol(A(:, ind)' * A(:, ind), 'lower');
        end
        xt = linsolve(CholFact, Aty(ind), linsolve_opts);
        xt = linsolve(CholFact, xt, linsolve_opts_transpose);
        
        res = y - A(:, ind) * xt;

        iter = 0;
        while numel(ind) < Kmax && iter < maxiter
            
            iter = iter+1;
            
            [~, ind_new] = sort(abs(A' * res), 'descend');
            ind_new = ind_new(1:alpha);
            
            ind_new = setdiff(ind_new, ind);
            
            % Update Cholesky factor
            if precompute_gram
                L21 = linsolve(CholFact, G(ind, ind_new), linsolve_opts);
                L21 = L21';
                L22 = G(ind_new, ind_new) - L21 * L21';                
            else
                L21 = linsolve(CholFact, A(:, ind)' * A(:, ind_new), ...
                    linsolve_opts);
                L21 = L21';
                L22 = A(:, ind_new)' * A(:, ind_new) - L21 * L21';
            end                
            L22 = chol(L22 + eps * eye(size(L22)), 'lower');
            CholFact = [CholFact, zeros(numel(ind), numel(ind_new)); ...
                L21, L22];
                                    
            ind = [ind; ind_new];
                        
            xt = linsolve(CholFact, Aty(ind(:)), linsolve_opts);
            xt = linsolve(CholFact, xt, linsolve_opts_transpose);
            
            [~, ind_rem] = sort(abs(xt));        
            ind_preserve = ind_rem(beta+1:end);
        
            ind = ind(ind_preserve);
        
            if precompute_gram
                CholFact = chol(G(ind, ind), 'lower');
            else
                CholFact = chol(A(:, ind)' * A(:, ind), 'lower');
            end
            xt = linsolve(CholFact, Aty(ind(:)), linsolve_opts);
            xt = linsolve(CholFact, xt, linsolve_opts_transpose);
            
            res = y - A(:, ind) * xt;
            
            if norm(res) < epsilon
                break
            end
            
            if verbose
                waitbar(iter / Kmax, hh)
            end
                
        end % while numel(ind) < Kmax
         
    case 'qr'                                              

        % Options for linsolve
        linsolve_opts.UT = true;                
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        [~, ind] = sort(abs(A' * res), 'descend');
        ind = ind(1:alpha);

        % Initial QR factors        
        [Qf, Rf] = qr(A(:, ind), 0);        

        % Update the approximation
        xt = linsolve(Rf, Qf' * y, linsolve_opts);

        [~, ind_rem] = sort(abs(xt));        
        ind_preserve = ind_rem(beta+1:end);
        
        ind = ind(ind_preserve);
                
        [Qf, Rf] = qr(A(:, ind), 0);
        
        xt = linsolve(Rf, Qf' * y, linsolve_opts);
        
        res = y - A(:, ind) * xt;

        iter = 0;
        while numel(ind) < Kmax && iter < maxiter
            
            iter = iter+1;
            
            [~, ind_new] = sort(abs(A' * res), 'descend');
            ind_new = ind_new(1:alpha);
            
            ind_new = setdiff(ind_new, ind);
            
            % Update QR factors
            R12 = Qf' * A(:, ind_new);                            
            [Q2, R22] = qr(A(:, ind_new) - Qf * R12, 0);             
            Qf = [Qf, Q2];
            
            Rf = [Rf, R12; 
                zeros(numel(ind_new), numel(ind)), R22];            
            
            %{
            [Qf, ~] = qr(Qf, 0);
            R2c = Qf' * A(:, ind_new);
            
            nind = numel(ind) + numel(ind_new);
            R = zeros(nind);
            R(1:numel(ind), 1:numel(ind)) = Rf;
            R(:, numel(ind)+1:end) = R2c;
            Rf = R;
            %}
                                    
            ind = [ind; ind_new];
                        
            xt = linsolve(Rf, Qf' * y, linsolve_opts);
            
            [~, ind_rem] = sort(abs(xt));        
            ind_preserve = ind_rem(beta+1:end);
        
            ind = ind(ind_preserve);
        
            [Qf, Rf] = qr(A(:, ind), 0);
            
            xt = linsolve(Rf, Qf' * y, linsolve_opts);
            
            res = y - A(:, ind) * xt;
            
            if norm(res) < epsilon
                break
            end
            
            if verbose
                waitbar(iter / Kmax, hh)
            end
            
        end % while numel(ind)       
        
    otherwise
        
        iter = 0;
        ind = [];
        
        while numel(ind) < Kmax && iter < maxiter
            
            iter = iter+1;
            
            [~, ind_new] = sort(abs(A' * res), 'descend');
            ind_new = ind_new(1:alpha);
            ind = union(ind, ind_new);
            
            xt = A(:, ind) \ y;
            
            [~, ind_rem] = sort(abs(xt));        
            ind_preserve = ind_rem(beta+1:end);
        
            ind = ind(ind_preserve);
            
            xt = A(:, ind) \ y;
            
            res = y - A(:, ind) * xt;
            
            if norm(res) < epsilon
                break
            end
            
            if verbose
                waitbar(iter / Kmax, hh)
            end
                
        end % while numel(ind) < Kmax
        
        
end

if verbose
    close(hh)
end

x = zeros(size(A, 2), 1);
x(ind) = xt;
