function [x, fv] = ISL0(A, y, varargin)
% Improved smoothed SL0 algorithm, from the paper: Md. M. Hyder, K. Mahata,
% An improved smoothed l_0 approximation algorithm for sparse
% representation, IEEE TrSP, 58(4), 2010.
%
% [x, fv] = ISL0(A, y, varargin)
%
% Input: A - mxn mixing matrix
%
%        y - mx1 right-side vector
%
%        optional arguments: sigmaMin - minimal value of sigma
%
%                            rho - sigma decrease factor
%
%                            eta - tolerance for stopping criterion
%
%                            gama - decrease factor in backtracking line
%                            search
%
% Output: x - nx1 solution vector
%
%         fv - function value at x
%
% M. F., 3.5.2010.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Some parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
A_pinv = pinv(A);
% Minimum l_2-norm solution
x = A_pinv * y;
% sigma0 = 2 * max(abs(x));

% Default values of optional parameters
sigmaMin = 1e-6;
rho = 0.5;
eta = 0.5;
gama = 0.5;
% verbose = false;

% Check for optional parameters
if(mod(length(varargin),2) == 1)
    error('Optional arguments come in pairs.');
end
for i = 1:2:length(varargin)
    switch lower(varargin{i})
        case 'sigmamin'
            sigmaMin = varargin{i+1};
        case 'rho'
            rho = varargin{i+1};
        case 'eta'
            eta = varargin{i+1};
        case 'gama'
            gama = varargin{i+1}; 
        case 'init'
            x = varargin{i+1};
        case 'sigmainit'
            sigma = varargin{i+1};
%         case 'verbose'
%             verbose = varargin{i+1};
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Main loop

if ~exist('sigma', 'var')
    sigma = 2 * max(abs(x));
end

while sigma >= sigmaMin
    kapa = 1;
    % Backtracking line search
    auxv = gfunc(x, A, y, sigma);
    xNew = kapa * auxv + (1-kapa)*x;
    while(ObjFunc(xNew, sigma) < ObjFunc(x, sigma))
        xNew = kapa * auxv + (1-kapa)*x;
        kapa = gama*kapa;
    end    
    xOld = x;
    x = xNew;
%     x = kapa * auxv + (1-kapa)*x;
    
    if norm(x-xOld) < eta*sigma
        sigma = rho * sigma;
    end
end

n = length(x);
fv = n - ObjFunc(x, sigmaMin);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function xOut = gfunc(xIn, A, y, sigma)
% Function used in algorithm
aux = 1e-10;
w = exp(-xIn.^2 / (2*sigma^2))+ aux;
AuxMat = bsxfun(@rdivide, A, w') * A';
AuxVect = AuxMat \ y;
xOut = (A' * AuxVect) ./ w;

function s = ObjFunc(xIn, sigma)
% Objective function
s = sum(exp(-xIn.^2 / (2*sigma^2)));
