function x = robust_SL0(A, y, sigmaNoise, varargin)
% Modified SL0 algorithm. Should be robust to noise.
%
% x = robust_SL0(A, y, sigmaNoise, varargin)
%
% Input: A - 'mixing' matrix
%
%        y - right side vector (y = Ax + noise)
%
%        sigmaNoise - approximate noise level
%
%        optional parameters: ...
%
% Input optional parameters as ('parameter name', 'parameter value') pairs
%
% Output: x - sparse solution
%
% [1] Eftekhari, Babaie-Zadeh, Jutten, Moghaddam, Robust-SL0 for stable
% sparse representations in noisy settings, ICASSP 2009.


% Parameters
inner_iters = 10;
sigmaMin = 1e-6;
decrease = 0.5;
mi = 1;
verbose = false;

if isfloat(A)
    D = @(x) A*x;       
    Apinv = pinv(A);
    Dpinv = @(x) Apinv*x;
elseif isa(A, 'function_handle') 
    D = A;    
else
    error('A is of wrong type!');
end

% Check for optional parameters
if(mod(length(varargin),2) == 1)
    error('Optional arguments come in pairs.');
end
for i = 1:2:length(varargin)
    switch lower(varargin{i})
        case 'apinv'
            Dpinv = varargin{i+1};
        case 'mi'
            mi = varargin{i+1};
        case 'inner'
            inner_iters = varargin{i+1};
        case 'decrease'
            decrease = varargin{i+1};
        case 'sigmamin'
            sigmaMin = varargin{i+1};
        case 'verbose'
            verbose = varargin{i+1};
    end
end

% Check if Apinv is provided
if ~exist('Dpinv', 'var')
    error('Pseudoinverse operator not provided!');
end

% % Estimate noise level
% aux = sigmaNoise * randn(m, 1);
% sigmaNoise = std(Dpinv(aux));

% Initialization
x = Dpinv(y);
sigma = 2 * max(max(abs(x)));
ind = 1:size(y, 2);

thresh = sqrt(size(A, 1)) * sigmaNoise;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Main algorithm

iter = 0;
maxiter = max(ceil(log(sigmaMin ./ sigma) ./ log(decrease)));

if verbose
    hh = waitbar(0, 'Algorithm in progress...');
end

while sigma > sigmaMin && ~isempty(ind)
    
    iter = iter + 1;
    
    for i = 1:inner_iters
                
        % Gradient descent step        
        delta = OurDelta(x(:, ind), sigma);
        x(:, ind) = x(:, ind) - mi * delta;
        
        residual = D(x) - y;
        
        % Projection step
        % Project if the norm of the residual is above the noise level
        % (this is the approach suggested in [1]).
        
        % Update ind
        ind = (sqrt(sum(residual.^2)) > thresh);
        
        x(:, ind) = x(:, ind) - Dpinv(residual(:, ind));
        
    end    
    
    sigma = sigma * decrease;
    
    if verbose
        waitbar(iter/maxiter, hh);
    end
    
end

if verbose
    close(hh)
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function delta = OurDelta(s, sigma)
% delta = s.*exp(-s.^2/sigma^2);
delta = s .* exp(-s.^2 / (2*sigma^2));

%{
function d = poly_der(s, gamma, sigma)
% Derivatative of polynomial approximation of l0, see eq.-s (13) and (16)
% in the paper: Mohimani, Babaie-Zadeh, Gorodnitsky, Jutten, Sparse
% recovery using SL0: Convergence analysis, arXiv

d = zeros(size(s));
Ind1 = (abs(s) <= 1);
Ind2 = (s > 1 & s <= (gamma+1));
Ind3 = (s < -1 & s >= (-gamma-1));
d(Ind1) = -2*s(Ind1) / (gamma+1);
d(Ind2) = 2*s(Ind2) / (gamma^2 + gamma) - 2/gamma;
d(Ind3) = 2*s(Ind3) / (gamma^2 + gamma) + 2/gamma;

d = -d;
%}
